Attribute VB_Name = "modPattern"
Option Explicit

' Copyright 2008 - Jason Peter Brown (jbrown@statslog.com)
' Free to use/modify in any way, under no warranty.

Private Declare Function GetSystemDefaultLCID Lib "kernel32" () As Long
Private Declare Function GetLocaleInfo Lib "kernel32" Alias "GetLocaleInfoA" (ByVal Locale As Long, ByVal LCType As Long, ByVal lpLCData As String, ByVal cchData As Long) As Long
Private Const LOCALE_SDECIMAL = &HE         '  decimal separator

Public Function PositiveNumericRangePattern(ByVal p_MinValue As Long, ByVal p_MaxValue As Long, ByVal p_MinDecimalPlaces As Long, ByVal p_MaxDecimalPlaces As Long, ByVal p_MaxValueInclusive As Boolean)
   Dim l_DecimalPlacesPattern                            As String
   Dim l_ValuesPattern                                  As String
   Dim l_MinDigits                                As Long
   Dim l_MaxDigits                                As Long
   Dim lo_Range As CMinMaxRange
   Dim lo_Ranges As Collection
   Dim i                                          As Long
   
      
   ' Validate Min/Max Value
   If p_MinValue > p_MaxValue Then
      Err.Raise 5    ' Invalid procedure call or argument
   ElseIf p_MinValue < 0 Or p_MaxValue < 0 Then
      ' Can't handle ranges < 0
      Err.Raise 5
   End If

   ' Validate Min/Max decimal places
   If p_MinDecimalPlaces < 0 Or p_MaxDecimalPlaces < 0 Then
      Err.Raise 5
   ElseIf p_MinDecimalPlaces > p_MaxDecimalPlaces Then
      Err.Raise 5
   ElseIf p_MinValue = p_MaxValue And _
          p_MaxDecimalPlaces > 0 Then
         ' Can't have decimal places if Min and Max values are the same
         Err.Raise 5
   ElseIf p_MinValue = p_MaxValue And _
          Not p_MaxValueInclusive Then
         ' Max value must be inclusive when Min and Max values are the same
         Err.Raise 5
   End If

   ' Generate regex for decimal places if required
   If p_MaxDecimalPlaces > 0 Then
      l_DecimalPlacesPattern = "(\" & DecimalCharacter & "\d{" & p_MinDecimalPlaces & "," & p_MaxDecimalPlaces & "}" & IIf(p_MinDecimalPlaces = 0, "|", "") & ")"
   End If

   ' Check for shortcuts
   If p_MinValue = p_MaxValue Then
      ' Min and Max values are the same, can be represented without complex regex pattern
      l_ValuesPattern = p_MinValue
      p_MaxValueInclusive = False ' Don't need to repeat the maximum value at the end of the pattern
   Else
      ' Get the number of digits in the Min and Max values
      l_MinDigits = IntegerDigitCount(p_MinValue)
      l_MaxDigits = IntegerDigitCount(p_MaxValue)
      
      If l_MinDigits = l_MaxDigits And _
         MostSignificantIntegers(p_MinValue, 1) = MostSignificantIntegers(p_MaxValue, 1) Then
         
         ' Min Value and Max value pattern can be calculated in 1 step
         ' Since they are the same digits in length and have the same
         ' most significant (left-most) digits
         
         ' Add range to collection for future processing
         Set lo_Ranges = New Collection
         Set lo_Range = New CMinMaxRange
         With lo_Range
            .Min = p_MinValue
            .Max = p_MaxValue - 1
         End With
         lo_Ranges.Add lo_Range
      
      Else
         ' Break min & max values into array of min/max powers of 10
         Set lo_Ranges = GetComplexPatternRanges(p_MinValue, p_MaxValue)
      End If
      
      ' Generate patterns to match each range
      For i = lo_Ranges.Count To 1 Step -1
         l_ValuesPattern = l_ValuesPattern & GetRangePattern(lo_Ranges(i).Min, lo_Ranges(i).Max)
         If i > 1 Then
            l_ValuesPattern = l_ValuesPattern & "|"
         End If
      Next i
      
      If LenB(l_DecimalPlacesPattern) <> 0 Then
         ' Surround values pattern in brackets to separate it from
         ' decimal places pattern if required
         l_ValuesPattern = "(" & l_ValuesPattern & ")"
      End If
   End If
   
   ' Compile complete regex pattern and return it
   PositiveNumericRangePattern = "^(" & l_ValuesPattern & l_DecimalPlacesPattern & IIf(p_MaxValueInclusive, "|" & p_MaxValue, "") & ")$"
End Function

Private Function DecimalCharacter() As String
   Static s_Char                                  As String
   Dim l_Len                                      As Long
   Dim l_Locale                                   As Long

   If LenB(s_Char) = 0 Then
      l_Locale = GetSystemDefaultLCID()
      l_Len = GetLocaleInfo(l_Locale, LOCALE_SDECIMAL, s_Char, Len(s_Char))
      If l_Len <> 0 Then
         s_Char = String$(l_Len, 0)

         GetLocaleInfo l_Locale, LOCALE_SDECIMAL, s_Char, Len(s_Char)

      End If
      s_Char = Left$(s_Char, InStr(1, s_Char, vbNullChar) - 1)
   End If

   DecimalCharacter = s_Char
End Function

Private Function GetRangePattern(ByVal p_Min As Long, ByVal p_Max As Long) As String
   Dim l_Pattern As String
   Dim l_MinDigits As Long
   Dim l_MaxDigits As Long
   Dim la_Min() As Long
   Dim la_Max() As Long
   Dim l_FullSubRanges As Long
   Dim i As Long
   Dim l_Digits As String
   
   l_MinDigits = IntegerDigitCount(p_Min)
   l_MaxDigits = IntegerDigitCount(p_Max)
   
   If l_MinDigits <> l_MaxDigits Then
      ' We can only parse ranges of the same number of digits
      Err.Raise 5
   End If
   
   la_Min = SplitPositiveInteger(p_Min)
   la_Max = SplitPositiveInteger(p_Max)
   For i = 1 To l_MinDigits
      If la_Min(i) = 0 And la_Max(i) = 9 Then
         ' Full sub-range can be represented with \d
         ' So mark it
         l_FullSubRanges = l_FullSubRanges + 1
      Else
         ' Not a full (i.e. 0-9) sub-range
         ' So collapse any previously encounter full sub-ranges
         l_Pattern = l_Pattern & CollapseFullSubRanges(l_FullSubRanges)
         ' The append the current non-full sub-range
         If la_Min(i) <> la_Max(i) Then
            If la_Max(i) - la_Max(i) = 1 Then
               ' consecutive digits
               l_Digits = la_Min(i) & la_Max(i)
            Else
               ' Range of digits
               l_Digits = la_Min(i) & "-" & la_Max(i)
            End If
            l_Pattern = l_Pattern & "[" & l_Digits & "]"
         Else
            ' Single digit representation
            l_Pattern = l_Pattern & la_Min(i)
         End If
      End If
   Next i

   ' Collapse any remaining full sub ranges into the pattern
   l_Pattern = l_Pattern & CollapseFullSubRanges(l_FullSubRanges)
   
   ' Return the regex pattern
   GetRangePattern = l_Pattern
End Function

Private Function CollapseFullSubRanges(ByRef p_FullSubRanges As Long) As String
   Dim l_Pattern As String
   
   If p_FullSubRanges > 0 Then
      l_Pattern = l_Pattern & "\d" & IIf(p_FullSubRanges > 1, "{" & p_FullSubRanges & "}", "")
      p_FullSubRanges = 0
   End If
   CollapseFullSubRanges = l_Pattern
End Function

Public Function Max(ByVal p_Num1 As Long, ByVal p_Num2 As Long) As Long
   ' Return the maximum of 2 passed integers
   
   If p_Num1 > p_Num2 Then
      p_Num2 = p_Num1
   End If

   Max = p_Num2
End Function

Public Function Min(ByVal p_Num1 As Long, ByVal p_Num2 As Long) As Long
   ' Return the minimum of 2 passed integers
   If p_Num1 > p_Num2 Then
      p_Num1 = p_Num2
   End If

   Min = p_Num1
End Function

Private Function SplitPositiveInteger(ByVal p_Num As Long) As Long()
   ' Splits a positive integer into an array of longs
   ' with each element holding one digit of the original integer
   
   Dim la_Num()                                   As Long
   Dim l_Digits                                   As Long
   Dim i                                          As Long

   If p_Num < 0 Then
      ' Can't process negative numbers
      Err.Raise 5
   End If

   l_Digits = IntegerDigitCount(p_Num)

   ReDim la_Num(1 To l_Digits)
   For i = l_Digits To 1 Step -1
      la_Num(i) = LeastSignificantIntegers(p_Num, 1)
      ' Drop least significant digit
      p_Num = p_Num \ 10
   Next i

   SplitPositiveInteger = la_Num()
End Function

Private Function GetComplexPatternRanges(ByVal p_Min As Long, ByVal p_Max As Long) As Collection
   ' Purpose: To convert a large range of numbers into management sub-ranges
   '          for parsing into regex patterns.
   ' Maybe the scariest code I have ever written. Sorry!
   
   Dim l_Min As Long
   Dim l_Max As Long
   Dim l_Lsi As Long
   Dim i As Long
   
   Dim lo_Ranges As Collection
   Dim lo_Range As CMinMaxRange
   
   Set lo_Ranges = New Collection
   
   l_Max = p_Max - 1
   l_Min = p_Min
   
   ' Get highest ranges by setting least significant digits to 0
   For i = 1 To IntegerDigitCount(p_Max - 1) - 1
      
      l_Lsi = LeastSignificantIntegers(p_Max, i)
      l_Min = p_Max - l_Lsi
      
      If l_Min >= p_Min Then
         If l_Min <= l_Max Then
         Set lo_Range = New CMinMaxRange
         With lo_Range
            .Min = l_Min
            .Max = l_Max
         End With
         lo_Ranges.Add lo_Range
         End If
         
         l_Max = l_Min - 1
      Else
         Exit For
      End If
   Next i
   
   ' Get each range from maximum digits to minimum digits
   If l_Min > p_Min Then
      For i = IntegerDigitCount(p_Max) - 1 To 1 Step -1
         l_Min = 10 ^ i
         If l_Min >= p_Min Then
            If l_Min <= l_Max Then
            Set lo_Range = New CMinMaxRange
            With lo_Range
               .Min = l_Min
               .Max = l_Max
            End With
            lo_Ranges.Add lo_Range
            End If
            
            l_Max = l_Min - 1
         Else
            Exit For
         End If
      Next i
   End If
   
   ' Get remaining ranges to minimum
   If l_Min <> p_Min Then
      l_Min = p_Min
      For i = IntegerDigitCount(p_Min) - 1 To 1 Step -1
         l_Min = Ceiling(p_Min / (10 ^ i)) * (10 ^ i)
         
         If l_Min <= l_Max Then
            Set lo_Range = New CMinMaxRange
            With lo_Range
               .Min = l_Min
               .Max = l_Max
            End With
            lo_Ranges.Add lo_Range
         End If
         
         l_Max = l_Min - 1
      Next i
   End If
   
   ' Create the minimum range
   Set lo_Range = New CMinMaxRange
   With lo_Range
      .Min = p_Min
      .Max = l_Max
   End With
   lo_Ranges.Add lo_Range
   

   ' Return the results
   Set GetComplexPatternRanges = lo_Ranges
End Function

Private Function Ceiling(ByVal p_Num As Double) As Long
   ' Returns the next highest integer from a real number (p_Num)
   Ceiling = -(Int(-p_Num))
End Function

Public Function LeastSignificantIntegers(ByVal p_Value As Long, ByVal p_Digits As Long) As Long
   ' Returns the right-most p_Digits integers from a passed long
   ' Example: LeastSignificantIntegers(1234, 1) = 4
   '          LeastSignificantIntegers(1234, 2) = 34
   
   LeastSignificantIntegers = p_Value Mod (10 ^ p_Digits)
End Function

Private Function MostSignificantIntegers(ByVal p_Num As Long, ByVal p_Digits As Long) As Long
   ' Returns the left-most p_Digits integers from a passed long
   ' Example: MostSignificantIntegers(1234, 1) = 1
   '          MostSignificantIntegers(1234, 2) = 12
   
   MostSignificantIntegers = p_Num \ (10 ^ (IntegerDigitCount(p_Num) - p_Digits))
End Function

Private Function IntegerDigitCount(ByVal p_Num As Long) As Long
   ' Returns the number of digits in an integer
   ' Example: IntegerDigitCount(10) = 2
   '          IntegerDigitCount(4567) = 4
   
   IntegerDigitCount = Len(CStr(Abs(p_Num)))
End Function

